from wikitools import comment_parser


class TestCommentParser:
    def test__inline_comments(self):
        for line, expected in (
            ("Empty", []),
            (
                "<!-- A single comment-->",
                [comment_parser.Comment(start=0, end=23)]
            ),
            (
                "<!-- Several --><!-- comments -->",
                [
                    comment_parser.Comment(start=0, end=15),
                    comment_parser.Comment(start=16, end=32),
                ]
            ),
        ):
            assert comment_parser.CommentParser().parse(line) == expected

    def test__multiline_comments(self):
        lines = [
            "I know my <!-- A",
            "B",
            "C",
            "and other --> letters.<!-- test -->",
            "Take a break",
            "<!-- and",
            "continue -->",
        ]

        comments = []
        parser = comment_parser.CommentParser()
        for line in lines:
            comments.extend(parser.parse(line))

        assert comments == [
            comment_parser.Comment(start=10, end=-1),
            comment_parser.Comment(start=-1, end=-1),
            comment_parser.Comment(start=-1, end=-1),
            comment_parser.Comment(start=-1, end=12),
            comment_parser.Comment(start=22, end=34),
            comment_parser.Comment(start=0, end=-1),
            comment_parser.Comment(start=-1, end=11),
        ]



from collections import Counter as multiset
import os

import tests.conftest
import tests.utils as utils

from wikitools import file_utils

class TestFileUtils:
    def test__list_all_article_files(self, root):
        article_paths = [
            'wiki/Article/en.md',
            'wiki/Article/fr.md',
            'wiki/Article/fil.md',
            'wiki/Article/pt-br.md',
            'wiki/Article/zh-tw.md',
            'wiki/Article/TRANSLATING.md',
            'wiki/Article/TEMPLATE.md',
            'wiki/Category1/Article/en.md',
            'wiki/Category1/Article/fr.md',
            'wiki/Category1/Article/pt-br.md',
            'wiki/Category1/Article/zh-tw.md',
            'wiki/Category1/Category2/Article/en.md',
            'wiki/Category1/Category2/Article/fr.md',
            'wiki/Category1/Category2/Article/pt-br.md',
            'wiki/Category1/Category2/Article/zh-tw.md',
            'wiki/Category1/Category2/Category3/Article/en.md',
            'wiki/Category1/Category2/Category3/Article/fr.md',
            'wiki/Category1/Category2/Category3/Article/pt-br.md',
            'wiki/Category1/Category2/Category3/Article/zh-tw.md',
        ]

        utils.create_files(root, *((path, '# Article') for path in article_paths))

        assert multiset(file_utils.list_all_articles()) == multiset(utils.remove(article_paths, "TEMPLATE.md", "TRANSLATING.md"))

    def test__list_all_article_dirs(self, root):
        article_paths = [
            'wiki/Article/en.md',
            'wiki/Article/fr.md',
            'wiki/Article/fil.md',
            'wiki/Article/pt-br.md',
            'wiki/Article/zh-tw.md',
            'wiki/Article/TRANSLATING.md',
            'wiki/Article/TEMPLATE.md',
            'wiki/Category1/Article/en.md',
            'wiki/Category1/Article/fr.md',
            'wiki/Category1/Article/pt-br.md',
            'wiki/Category1/Article/zh-tw.md',
            'wiki/Category1/Category2/Article/en.md',
            'wiki/Category1/Category2/Article/fr.md',
            'wiki/Category1/Category2/Article/pt-br.md',
            'wiki/Category1/Category2/Article/zh-tw.md',
            'wiki/Category1/Category2/Category3/Article/en.md',
            'wiki/Category1/Category2/Category3/Article/fr.md',
            'wiki/Category1/Category2/Category3/Article/pt-br.md',
            'wiki/Category1/Category2/Category3/Article/zh-tw.md',
        ]

        utils.create_files(root, *((path, '# Article') for path in article_paths))

        assert multiset(file_utils.list_all_article_dirs()) == multiset(set(os.path.dirname(path) for path in article_paths))

    def test__list_all_translations(self, root):
        article_paths = [
            'wiki/Article/en.md',
            'wiki/Article/fr.md',
            'wiki/Article/fil.md',
            'wiki/Article/pt-br.md',
            'wiki/Article/zh-tw.md',
            'wiki/Article/TRANSLATING.md',
            'wiki/Article/TEMPLATE.md',
        ]

        utils.create_files(root, *((path, '# Article') for path in article_paths))

        assert multiset(file_utils.list_all_translations(["wiki/Article"])) == multiset(article_paths[1:5])

    def test_list_all_articles_and_newsposts(self, root):
        article_paths = [
            'wiki/Article/en.md',
            'wiki/Article/fil.md',
            'wiki/Article/fr.md',
            'wiki/Article/pt-br.md',
            'wiki/Article/zh-tw.md',
            'wiki/Article/TRANSLATING.md',
            'wiki/Article/TEMPLATE.md',
        ]

        newspost_paths = [
            'news/2022-06-23-project-loved-june-2022.md',
            'news/2023-09-22-first-world-cup-held-using-lazer.md',
            'news/2024-05-15-2b-maps-are-now-rankable.md',
            'news/2025-12-20-trivium-quiz-winners.md',
            'news/2026-03-12-second-annual-pp-committee-meeting.md',
            'news/2027-02-08-the-old-client-is-now-deprecated.md',
            'news/2028-07-28-introducing-osu-lite-smartwatch-edition.md',
            '.remarkrc.js'
        ]

        utils.create_files(root, *((path, '# Article') for path in article_paths))
        utils.create_files(root, *((path, '# News post') for path in newspost_paths))

        assert multiset(file_utils.list_all_articles_and_newsposts()) == multiset(article_paths[0:5] + newspost_paths[0:-1])



from wikitools import code_block_parser


class TestCodeBlockParser:
    def test__inline_blocks(self):
        for line, expected in (
            ("Empty", []),
            (
                "`test`",
                [code_block_parser.CodeBlock(start=0, end=5)]
            ),
            (
                "`several` `code blocks`",
                [
                    code_block_parser.CodeBlock(start=0, end=8),
                    code_block_parser.CodeBlock(start=10, end=22),
                ]
            ),
            (
                "``block-o``",
                [code_block_parser.CodeBlock(start=0, end=10)]
            ),
            (
                "`` `Space` ``",
                [code_block_parser.CodeBlock(start=0, end=12)]
            ),
            (
                "`` code block with random backticks ` ``` ` ``` ``",
                [code_block_parser.CodeBlock(start=0, end=49)]
            ),
            (
                "stray backtick ` and a ``code block`` and another stray ```",
                [code_block_parser.CodeBlock(start=23, end=36)]
            ),
        ):
            parser = code_block_parser.CodeBlockParser()
            assert parser.parse(line) == expected, line
            assert not parser.in_multiline

    def test__multiline_blocks(self):
        lines = [
            "This is expected:",
            "```",
            "Valid code block",
            "```",
            "````markdown",
            "```",
            "Code block inside code block (ignored)",
            "Inline `code blocks` inside a code block",
            "```",
            "````",
            "This won't start a multi-line code block: ```",
            "Here we have a `valid in-line code block`",
            "```",
            "Trailing multi-line code block",
        ]

        blocks = []
        parser = code_block_parser.CodeBlockParser()
        for line in lines:
            blocks.extend(parser.parse(line))

        assert blocks == [
            *(9 * [code_block_parser.CodeBlock(start=-1, end=-1)]),

            code_block_parser.CodeBlock(start=15, end=40),
            code_block_parser.CodeBlock(start=-1, end=-1),
            code_block_parser.CodeBlock(start=-1, end=-1),
        ]
        assert parser.in_multiline



from urllib import parse

from wikitools import link_parser, reference_parser


class TestInlinePlainLinks:
    def test__regular_link(self):
        example = "An [example](/wiki/Example)."
        link = link_parser.find_link(example)
        assert link == link_parser.Link(
            start=3,
            end=26,
            alt_text="example",
            raw_location="/wiki/Example",
            parsed_location=parse.urlparse("/wiki/Example"),
            title="",
            is_reference=False,
        )

    def test__regular_link__title(self):
        example = 'An [example](/wiki/Example "Title").'
        link = link_parser.find_link(example)
        assert link == link_parser.Link(
            start=3,
            end=34,
            alt_text="example",
            raw_location="/wiki/Example",
            parsed_location=parse.urlparse("/wiki/Example"),
            title=' "Title"',
            is_reference=False,
        )

    def test__no_link(self):
        example = 'Check this out: [[a]].'
        assert link_parser.find_links(example) == []

    def test__no_link__in_square_brackets(self):
        example = 'Check this out: [[a](/wiki/A)].'
        assert link_parser.find_links(example) == []

    def test__link__nested_brackets(self):
        example = 'Check this out: [a](/wiki/[A])].'
        assert link_parser.find_links(example) == [
            link_parser.Link(
                start=16,
                end=29,
                alt_text="a",
                raw_location="/wiki/[A]",
                parsed_location=parse.urlparse("/wiki/[A]"),
                title="",
                is_reference=False,
            )
        ]

    def test__link__fragment(self):
        example = "RTF[M](/wiki/M#manual)."
        assert link_parser.find_link(example) == link_parser.Link(
            start=3,
            end=21,
            alt_text="M",
            raw_location="/wiki/M#manual",
            parsed_location=parse.urlparse("/wiki/M#manual"),
            title="",
            is_reference=False,
        )

    def test__link__query_string(self):
        example = "[example.com](https://example.com/?test=1) for sale!"
        assert link_parser.find_link(example) == link_parser.Link(
            start=0,
            end=41,
            alt_text="example.com",
            raw_location="https://example.com/?test=1",
            parsed_location=parse.urlparse("https://example.com/?test=1"),
            title="",
            is_reference=False,
        )


class TestInlineImageLinks:
    def test__image_link(self):
        example = "Check this out: ![](/wiki/crown.png)"
        link = link_parser.find_link(example)
        assert link == link_parser.Link(
            start=17,
            end=35,
            alt_text="",
            raw_location="/wiki/crown.png",
            parsed_location=parse.urlparse("/wiki/crown.png"),
            title="",
            is_reference=False,
        )

    def test__image_link__alt_text(self):
        example = "Check this out: ![Crown](/wiki/crown.png)"
        link = link_parser.find_link(example)
        assert link == link_parser.Link(
            start=17,
            end=40,
            alt_text="Crown",
            raw_location="/wiki/crown.png",
            parsed_location=parse.urlparse("/wiki/crown.png"),
            title="",
            is_reference=False,
        )

    def test__image_link__title(self):
        example = 'Check this out: ![](/wiki/crown.png "Title")'
        link = link_parser.find_link(example)
        assert link == link_parser.Link(
            start=17,
            end=43,
            alt_text="",
            raw_location="/wiki/crown.png",
            parsed_location=parse.urlparse("/wiki/crown.png"),
            title=' "Title"',
            is_reference=False,
        )

    def test__image_link__alt_text__title(self):
        example = 'Check this out: ![Crown](/wiki/crown.png "Title")'
        link = link_parser.find_link(example)
        assert link == link_parser.Link(
            start=17,
            end=48,
            alt_text="Crown",
            raw_location="/wiki/crown.png",
            parsed_location=parse.urlparse("/wiki/crown.png"),
            title=' "Title"',
            is_reference=False,
        )


class TestInlineMultipleLinks:
    def test__multiple_links(self):
        example = 'Check this out: [a](/wiki/A) and [b](/wiki/B).'
        links = link_parser.find_links(example)
        assert links == [
            link_parser.Link(
                start=16,
                end=27,
                alt_text="a",
                raw_location="/wiki/A",
                parsed_location=parse.urlparse("/wiki/A"),
                title="",
                is_reference=False,
            ),
            link_parser.Link(
                start=33,
                end=44,
                alt_text="b",
                raw_location="/wiki/B",
                parsed_location=parse.urlparse("/wiki/B"),
                title="",
                is_reference=False,
            )
        ]


class TestReferenceLinks:
    def test__regular_link(self):
        example = "See for [yourself][reference]."
        link = link_parser.find_link(example)
        assert link == link_parser.Link(
            start=8,
            end=28,
            alt_text="yourself",
            raw_location="reference",
            parsed_location=parse.urlparse("reference"),
            title="",
            is_reference=True,
        )

    def test__image_link(self):
        example = "No crowns here: ![Sweden][SE_flag]"
        link = link_parser.find_link(example)
        assert link == link_parser.Link(
            start=17,
            end=33,
            alt_text="Sweden",
            raw_location="SE_flag",
            parsed_location=parse.urlparse("SE_flag"),
            title="",
            is_reference=True,
        )


class TestIgnoredFootnotes:
    def test__ignored_footnotes(self):
            examples = [
                "The sky is blue.[^reference-1]",
                "The sky is blue.[^reference-1][^reference-2]",
                "The sky is blue.[^reference-1][^reference-2][^reference-3]",
            ]
            for example in examples:
                assert link_parser.find_link(example) == None
                assert link_parser.find_links(example) == []

    def test__footnote_after_link(self):
        example = "[test](https://www.example.com/)[^some-footnote]"
        link = link_parser.find_link(example)
        assert link == link_parser.Link(
            start=0,
            end=31,
            alt_text="test",
            raw_location="https://www.example.com/",
            parsed_location=parse.urlparse("https://www.example.com/"),
            title="",
            is_reference=False,
        )
        assert len(link_parser.find_links(example)) == 1

    def test__footnote_after_reference_link(self):
        example = "[test][https://www.example.com/][^some-footnote]"
        link = link_parser.find_link(example)
        assert link == link_parser.Link(
            start=0,
            end=31,
            alt_text="test",
            raw_location="https://www.example.com/",
            parsed_location=parse.urlparse("https://www.example.com/"),
            title="",
            is_reference=True,
        )
        assert len(link_parser.find_links(example)) == 1


class TestIdentifierLinks:
    def test__misc_links(self):
        for example, (start, fragment_start) in (
            ("Thank you, we'll [call](/wiki/Test#test) your name", (17, 34)),
            ("I am sure you [will](#local-ref).", (14, 21)),
            ("No [reference](/wiki/No).", (3, 23)),
        ):
            link = link_parser.find_link(example)
            assert link
            assert link.start == start
            assert link.fragment_start == fragment_start
            print(link.end)


class TestLinkObject:
    def test__resolution(self):
        for link, reference, (raw_location, parsed_location) in (
            ('This is an [example][example_ref].', '[example_ref]: https://example.com "Example"', ('https://example.com', parse.urlparse("https://example.com"))),
            ('Cats are cute. [[1]][r]', '[r]: #references', ('#references', parse.urlparse("#references"))),
        ):
            link = link_parser.find_link(link)
            assert link
            reference = reference_parser.extract(reference, lineno=1)
            assert reference

            resolved_reference = link.resolve({reference.name: reference})
            assert resolved_reference
            assert resolved_reference.raw_location == reference.raw_location
            assert resolved_reference.raw_location == raw_location
            assert resolved_reference.parsed_location == reference.parsed_location
            assert resolved_reference.parsed_location == parsed_location
            assert link.resolve({}) is None
